import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.net.URL;
import java.util.Random;

import javax.imageio.ImageIO;


public class CardDeck implements Cloneable{

	public enum SUIT { Club, Spade, Heart, Diamond }

	public enum SUIT_COLOR { Black, White }

	public static final SUIT_COLOR[] cardColor = { SUIT_COLOR.Black, SUIT_COLOR.Black, SUIT_COLOR.White, SUIT_COLOR.White };
	
	public enum FACE { Ace, King, Queen, Jack, Ten, Nine, Eight,
		Seven, Six, Five, Four, Three, Two 	}
	
	public static final int NUM_CARDS_IN_DECK = 52;
	
	public static int CARD_WIDTH = 0;
	public static int CARD_HEIGHT = 0;
	
	public static final String imageExtension = ".png";
	
	Card[] deck;
	
	/**
	 * This is used solely for testing a specific layout
	 */
	public CardDeck(URL codeBase, String filePath) throws Exception {
		// create the array to hold the card deck
		deck = new Card[NUM_CARDS_IN_DECK];
		
		BufferedReader br = new BufferedReader(new FileReader(filePath));
		String s;
		int nextAvailableSlot = 0;
		while((s = br.readLine()) != null) {
			nextAvailableSlot = processLine(s, nextAvailableSlot, codeBase);
		} 
		br.close(); 
		setCardDimensions();		
	}
	
	public CardDeck(Card[] newDeck) {
		deck = newDeck;
	}
	
	public Object clone() throws CloneNotSupportedException {
		if (deck == null) return null;
		Card[] deckCopy = new Card[deck.length];
		for (int i=0; i < deck.length; i++) {
			deckCopy[i] = (Card) deck[i].clone();
		}
		return new CardDeck(deckCopy);
	}
	
	/**
	 * Prints the card deck
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("CARD DECK\n");
		for (int i=0; i < deck.length; i++) {
			sb.append("Card " + String.valueOf(i) + ": "  + deck[i].toString() + "\n");
		}
		return sb.toString();
	}
	/** Constructor expects to find in resources/images the image for each card
	 *  Each image is read and saved in the deck array
	 */
	public CardDeck(URL codeBase) {
		// create the array to hold the card deck
		deck = new Card[NUM_CARDS_IN_DECK];
		
		int cardNum = 1;	// iterates through the value 1..52 since the cards file names are in 
							// that order
		// populate the deck of cards
		for (CardDeck.FACE faceValue : CardDeck.FACE.values()) {
			for (CardDeck.SUIT suitValue : CardDeck.SUIT.values()) {
				// read in cardNum.png and create card with matching FACE and SUIT value
				BufferedImage img = null;
				try {
					//URL url = new URL(codeBase, FreeCell.imagePath + String.valueOf(cardNum) + imageExtension);
					img = ImageIO.read(getClass().getResourceAsStream(FreeCell.imagePath + String.valueOf(cardNum) + imageExtension));
					Card aCard = new Card(suitValue, faceValue, img);
					deck[cardNum-1] = aCard;
					cardNum++;
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		setCardDimensions();
	}

	private void setCardDimensions() {
		// set the CARD_LENGTH and CARD_WIDTH
		if (deck != null && deck.length > 0) {
			CARD_WIDTH = deck[0].getImage().getWidth();
			CARD_HEIGHT = deck[0].getImage().getHeight();
		} 		
	}
	public Card getCard(int i)  {
		if ( i < 0 || i >= NUM_CARDS_IN_DECK) {
			System.out.println("Card slot is out of range 0..51 expected.");
			return null;
		}
		return deck[i];
	}
	
	private CardDeck.FACE getCardFace(String strFace) {
		if (strFace.equalsIgnoreCase("A")) {
			return CardDeck.FACE.Ace;
		} else if (strFace.equalsIgnoreCase("K")) {
			return CardDeck.FACE.King;
		} else if (strFace.equalsIgnoreCase("Q")) {
			return CardDeck.FACE.Queen;
		} else if (strFace.equalsIgnoreCase("J")) {
			return CardDeck.FACE.Jack;
		} else {
			return CardDeck.FACE.values()[14-Integer.parseInt(strFace)];
		}
		
	}
	
	private CardDeck.SUIT getCardSuit(String strSuit) {
		if (strSuit.equalsIgnoreCase("C")) {
			return CardDeck.SUIT.Club;
		} else if (strSuit.equalsIgnoreCase("S")) {
			return CardDeck.SUIT.Spade;
		} else if (strSuit.equalsIgnoreCase("H")) {
			return CardDeck.SUIT.Heart;
		} else {
			return CardDeck.SUIT.Diamond;
		}
	}
	
	private int determineCardNum(CardDeck.FACE cardFace, CardDeck.SUIT cardSuit) {
		return cardFace.ordinal() * 4 + cardSuit.ordinal() + 1;
	}
	
	private Card mapCard(String strFace, String strSuit, URL codeBase, boolean over10) throws Exception {
		CardDeck.FACE cardFace;
		CardDeck.SUIT cardSuit;
		BufferedImage img;
		// Determine card number
		cardFace = getCardFace(strFace);
		cardSuit = getCardSuit(strSuit);
		int cardNum = determineCardNum(cardFace, cardSuit);
		//URL url = new URL(codeBase, FreeCell.imagePath + String.valueOf(cardNum) + imageExtension);
		img = ImageIO.read(getClass().getResourceAsStream(FreeCell.imagePath + String.valueOf(cardNum) + imageExtension));
		Card aCard = new Card(cardSuit, cardFace, img);
		return aCard;
	}
	
	private int processLine(String s, int nextAvailableSlot, URL codeBase) throws Exception{
		String[] cardStr = s.split("\\s+");
		String strFace = null;
		String strSuit = null;
		for (String aCardDesignation:cardStr) {
			Card card = null;
			if (aCardDesignation.matches("^[AKQJ][CDHS)]$")) {
				strFace = aCardDesignation.substring(0, 1);
				strSuit = aCardDesignation.substring(1);
				card = mapCard(strFace, strSuit, codeBase, true);
				deck[nextAvailableSlot++] = card;
			} else {
				// extract face and suit from digit card
				if (aCardDesignation.startsWith("1") && aCardDesignation.length() == 3) {
					strFace = aCardDesignation.substring(0, 2);
					strSuit = aCardDesignation.substring(2);	
				} else if (aCardDesignation.length() == 2){
					strFace = aCardDesignation.substring(0,1);
					strSuit = aCardDesignation.substring(1);
				}
				card = mapCard(strFace, strSuit, codeBase, false);
				deck[nextAvailableSlot++] = card;
			}
		}
		return nextAvailableSlot;
	}
	public void shuffle() {
		// simulate a shuffle by randomly moving each card
		Random randomGenerator = new Random();
		int randomIterations = randomGenerator.nextInt(5)+1;
		System.out.println("Shuffling deck " + randomIterations + " times.");
		for (int i=0; i <= randomIterations; i++) {
			for (int cardIndex = 0; cardIndex < NUM_CARDS_IN_DECK; cardIndex++) {
				int randomPosition = randomGenerator.nextInt(NUM_CARDS_IN_DECK);
				// swap with cardIndex Position
				Card tempCard = deck[cardIndex];
				deck[cardIndex] = deck[randomPosition];
				deck[randomPosition] = tempCard;
			}
		}
		System.out.println("Done shuffling deck.");
	}
	

}
