import java.awt.Graphics2D;
import java.awt.Point;
import java.util.ArrayList;


public interface CardPlayableArea {

	/* Space between cards in a play field area */
	public static final int CARD_GAP = 10;	
	
	/* Round rectange arc used when drawing available space */
	public static final int CARD_ROUND_ARC = 10;

	/**
	 * This is used to determine if mouse is in the area screen section or not
	 * 
	 * @param point
	 * @return
	 */
	public boolean isPointInArea(Point point);
	/**
	 * Determines if the give point is under a valid card in this play area
	 * 
	 * @param point
	 * @return 	boolean if mouse pointer or release was at a card position in play area
	 */
	public boolean isCardHere(Point point);
	
	/**
	 * Returns the card (or cards) from this point in the play area. Note: For CardPlayField it may be a row of cards
	 * @param point
	 * @return
	 */
	public ArrayList<Card> grabCards(Point point, boolean removeFromList);
	
	
	/**
	 * Returns the number of available slots in the given play card area
	 * @return
	 */
	public int numAvailableSlots();
	
	/**
	 * Adds card to this location in the play field if allowed to
	 * @param card
	 * @param point
	 * @return
	 */
	public boolean dropCard(ArrayList<Card> cards, Point point, boolean validate);
	
	/** Display/renders the play field
	 * 
	 * @param g2D
	 */
	public void render(Graphics2D g2D);
	
	/**
	 * This method returns a card that qualifies for auto-movement to the foundation.
	 * @return
	 */
	public Card findCardToAutoMove(CardAutoAddable autoAddableArea);
	
	/**
	 * This method returns the card in this area (or null)
	 * 
	 * @param point
	 * @return
	 */
	public Card getCardHere(Point point);
	
}
