import java.awt.AWTEvent;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
//import java.net.URL;
import java.util.ArrayList;

import javax.imageio.ImageIO;
import javax.swing.JApplet;


public class FreeCell extends JApplet 
	implements Runnable, MouseListener, MouseMotionListener, KeyListener, EventProcessable {

	private static final long serialVersionUID = -7700295997049184899L;
	private static final String congratulationsMessage = "Congrats you have completed the game!";
	private static final String appletVersionNo = "Version #0.82";
	public static final int DISPLAY_WIDTH = 740;
	public static final int DISPLAY_HEIGHT = 480;
	public static final String imagePath = "images/";
	
	private BufferedImage	backBuffer;
	private Graphics2D		bbGraphics;
	private BufferedImage	mainScreenImage;
	
	private	Thread loop;	// thread running our game loop
	
	private EventProcessor eventProcessor;	// Holds all key presses and mouse events
	
	private CardDeck deck;
	private CardDeck deckBackup;	// used to re-start game if user gets stuck
	private Card cheatCard;			// this saves the current card the user wants to look at fully in order
									// to see what suit it is in
	
	CardPlayField cardPlayField;	// represents our three card fields: foundation, open area and tableau
	
	// What occurs on the screen is determined by the gameState:
	// WAITING_FOR_USER_INPUT 	- waiting for user to make move
	// SOURCE_CARD_SELECTED		- user has clicked on a card (it will be highlighted with halo effect)
	// GAME_DONE				- we are in final animation state
	// CARD_BEING_GRABBED		- user is dragging a card from one location to another
	// ANIMATE_CARD_RETURN		- user incorrectly dragged card and released in a wrong spot
	//								card will be animated back to its original spot
	// ANIMATE_CARD_AUTO_MOVE	- card has been selected for auto-move and is being animated from
	//								open area or tableau to the foundation
	// REVEAL_CARD				- when the user right-clicks on a card, that card is "fully" displayed to the user
	// GAME_START				- game is displaying the start screen, user must press "S" to start the game
	public enum GAME_STATE { WAITING_FOR_USER_INPUT, SOURCE_CARD_SELECTED, GAME_DONE, CARD_BEING_GRABBED,
		ANIMATE_CARD_RETURN, ANIMATE_CARD_AUTO_MOVE, GAME_START, REVEAL_CARD}
	

	GAME_STATE gameState;
	MovingCard sourceMovingCard;	// populated when user either clicks (select source), drags or double-clicks a card
	boolean	canGameAutoMove = false; // turned on after the user makes the first move in the game
	
	// Initialize the game parameters
	public void init() {
		// Display information
		getContentPane().setLayout(null);
		setSize(DISPLAY_WIDTH, DISPLAY_HEIGHT);
		setIgnoreRepaint(true);
		
		// Set up the back buffer
		backBuffer = new BufferedImage(DISPLAY_WIDTH, DISPLAY_HEIGHT, BufferedImage.TYPE_INT_RGB);
		bbGraphics = (Graphics2D)backBuffer.getGraphics();		
		
		setUpMainScreen();
		
		
		gameState = GAME_STATE.GAME_START;	// This determines what the game is showing the user
		
		eventProcessor = new EventProcessor(this);
		
		addMouseListener(this);
		addMouseMotionListener(this);
		addKeyListener(this);
		setFocusable(true);
	}
	


	private void setUpMainScreen()  {
		try {
			mainScreenImage = ImageIO.read(getClass().getResourceAsStream(imagePath + "cooltext715146918.png"));		
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	// Called if the user is stuck in current game or completed the game and wants to start over
	private void startNewGame() {
		canGameAutoMove = false;
		sourceMovingCard = null;
		cheatCard = null;
		try {
			deck = (CardDeck) deckBackup.clone();
			deck.shuffle();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		cardPlayField = new CardPlayField(deck);
		gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
		
	}
	
	private  void setUpCardDeck() {
		try {
			deck = new CardDeck(getCodeBase());
			deck.shuffle();
			// TODO: Remove test code		
			//deck = new CardDeck(getCodeBase(), "../resources/data/testcase03.txt");
			deckBackup = (CardDeck) deck.clone();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		cardPlayField = new CardPlayField(deck);
		gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
	}

	private void reStartGame() {
		canGameAutoMove = false;
		cardPlayField = new CardPlayField(deckBackup);
		gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
		
	}
	
	public void start() {
		loop = new Thread(this);
		loop.start();
		requestFocusInWindow();
	}
	
	public void stop() {
		loop = null;
	}
	public void render(Graphics2D g2D) {
		// Make the background green
		Color darkGreenColor = new Color(71, 131,95);
		g2D.setColor(darkGreenColor);
		g2D.fillRect(0, 0, DISPLAY_WIDTH, DISPLAY_HEIGHT);
				
		// Make the background green
		if (cardPlayField != null) {
			cardPlayField.render(g2D);
		}

		switch (gameState) {
			case SOURCE_CARD_SELECTED:
			case CARD_BEING_GRABBED:
			case ANIMATE_CARD_RETURN:
			case ANIMATE_CARD_AUTO_MOVE:
				sourceMovingCard.render(g2D);
				break;
			case GAME_DONE:
				g2D.setColor(Color.black);
				drawCongratulationsMessage(g2D);
				break;
			case GAME_START:
				// No, display front screen
				int x = (DISPLAY_WIDTH - mainScreenImage.getWidth()) / 2;
				int y = (DISPLAY_HEIGHT - mainScreenImage.getHeight()) / 2;
				g2D.drawImage(mainScreenImage, null, x, y);
				g2D.setColor(Color.white);
				g2D.drawString(appletVersionNo, 10, 30);
				break;
			case REVEAL_CARD:
				// print the reveal card on it's  own
				cheatCard.render(g2D);
				break;
		}

	}

	private void drawCongratulationsMessage(Graphics2D g2D) {
		Font oldFont = g2D.getFont();
		Font newFont = new Font("Verdana", Font.BOLD, 24);
		g2D.setFont(newFont);
		FontMetrics fm = g2D.getFontMetrics();
		Rectangle2D rect2D = fm.getStringBounds(congratulationsMessage, g2D);
		
		int textHeight = (int)(rect2D.getHeight());
		int textWidth = (int) (rect2D.getWidth());
	
		int x = (DISPLAY_WIDTH - textWidth) / 2;
		int y = (DISPLAY_HEIGHT - textHeight) /2 + fm.getAscent();
		
		g2D.drawString(congratulationsMessage, x, y);
		g2D.setFont(oldFont);
	}
	
	@Override
	public void mouseClicked(MouseEvent event) {
		eventProcessor.addEvent(event);
	}

	@Override
	public void mouseEntered(MouseEvent event) {
		//eventProcessor.addEvent(event);
	}

	@Override
	public void mouseExited(MouseEvent event) {
		//eventProcessor.addEvent(event);	
	}

	@Override
	public void mousePressed(MouseEvent event) {
		eventProcessor.addEvent(event);
	}

	@Override
	public void mouseReleased(MouseEvent event) {
		eventProcessor.addEvent(event);
	}
	
	@Override
	public void mouseDragged(MouseEvent event) {
		eventProcessor.addEvent(event);
	}

	@Override
	public void mouseMoved(MouseEvent event) {
		//eventProcessor.addEvent(event);
	}
	
	@Override
	public void keyPressed(KeyEvent event) {
		eventProcessor.addEvent(event);
		
	}

	@Override
	public void keyReleased(KeyEvent event) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void keyTyped(KeyEvent event) {
		// TODO Auto-generated method stub		
	}
	
	@Override
	public void run() {
		long startTime;
		long waitTime;
		long elapsedTime;
		// 1000 / 25 Frames Per Second = 40 milliseconds delay
		int delayTime = 1000 / 25;
		
		Thread thisThread = Thread.currentThread();
		while (loop == thisThread) {
			startTime = System.currentTimeMillis();
			
			// handle all events first
			eventProcessor.processEventList();
			
			// do any animation here (returning cards back . . . )
			if (gameState == GAME_STATE.ANIMATE_CARD_RETURN || gameState == GAME_STATE.ANIMATE_CARD_AUTO_MOVE) {
				if (sourceMovingCard.moveTowardsOriginalPosition()) {
					// done with animation add card back to original list
					cardPlayField.returnCards(sourceMovingCard.getCardList(), sourceMovingCard.getOriginalCardPoint());
					gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
				}
			}
			// auto animate any cards that can go up to the pile
			if (gameState == GAME_STATE.WAITING_FOR_USER_INPUT && canGameAutoMove) {
				checkAutoMoveCards();
			}
			
			// render to back buffer now
			render(bbGraphics);
			
			// render back buffer image to screen
			Graphics g = getGraphics();
			if (g != null) {
				g.drawImage(backBuffer, 0, 0, null);
				g.dispose();
			}
			
			// handle frame rate
			elapsedTime = System.currentTimeMillis() - startTime;
			waitTime = Math.max(delayTime - elapsedTime, 5);
			
			try {
				Thread.sleep(waitTime);
			} catch (InterruptedException e) {
				// ignore
			}
			
			if (cardPlayField != null && cardPlayField.isGameOver()) {
				gameState = GAME_STATE.GAME_DONE;
				
			}
			
		}
	}

	// this class checks if any cards can be auto-moved. If found we simulate a double-click on the card.
	private void checkAutoMoveCards() {
		Card card = null;
		ArrayList<Card> cardList = null;
		card = cardPlayField.autoMoveAnyCards();
		
		if (card != null) {
			int x = card.getScreenPosition().x + (CardDeck.CARD_WIDTH / 2);
			int y = card.getScreenPosition().y + (CardDeck.CARD_HEIGHT / 2);
			Point mePoint = new Point(x, y);	// points to the middle of the card
			if (cardPlayField.isSingleClickInValidArea(mePoint) && cardPlayField.isCardUnderPoint(mePoint)) {
				cardList = cardPlayField.pickUpCards(mePoint, true); 
				if (cardList != null && cardList.size() == 1) {
					Point theTargetPoint = cardPlayField.determineCardMoveArea(cardList);
					if (theTargetPoint != null) {
						Point oldCardPoint = cardList.get(0).getScreenPosition();
						cardList.get(0).setScreenPosition(theTargetPoint);
						sourceMovingCard = new MovingCard(cardList, oldCardPoint, oldCardPoint);
						sourceMovingCard.markAnimationSpot();
						canGameAutoMove = true;
						gameState = GAME_STATE.ANIMATE_CARD_AUTO_MOVE;
					} else {
						cardPlayField.returnCards(cardList, mePoint);
						gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
					}
				} else {
					gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
				}
			}
				
		}		
	}
	
	@Override
	public void handleEvent(AWTEvent event) {
		MouseEvent me = null;
		KeyEvent ke = null;
		ArrayList<Card> cardList = null;
		Point mePoint = null;
		switch(event.getID()) {
		
			case KeyEvent.KEY_PRESSED:	// User has request start or re-start of the game
				ke = (KeyEvent) event;
				int keyCode = ke.getKeyCode();
				if (keyCode == KeyEvent.VK_S) {
					System.out.println("User clicked on GAME_START key");
					if (gameState == GAME_STATE.GAME_START) { 
						setUpCardDeck();
					} else {
						reStartGame();
					}
				}
				if (keyCode == KeyEvent.VK_N) {
					// The user is requesting a new game
					System.out.println("User clicked on GAME_START key");
					startNewGame();
				}
				break;
			case MouseEvent.MOUSE_PRESSED: 
				me = (MouseEvent) event;
				mePoint = me.getPoint();
				if (me.getButton() == MouseEvent.BUTTON3 && gameState == GAME_STATE.WAITING_FOR_USER_INPUT) {
					// check if user under card, if so remember card to display front and center
					cheatCard = cardPlayField.getValidPeakCard(mePoint);
					if (cheatCard != null) {
						gameState = GAME_STATE.REVEAL_CARD;
					}
				}
				break;
				
			case MouseEvent.MOUSE_DRAGGED:	// User may be dragging a card
				me = (MouseEvent) event;
				mePoint = me.getPoint();
				if (gameState == GAME_STATE.CARD_BEING_GRABBED) {
					// get current position position and update moving card
					sourceMovingCard.setCurrentPoint(new Point (mePoint.x - sourceMovingCard.getMouseOffsetPoint().x,
							mePoint.y - sourceMovingCard.getMouseOffsetPoint().y));
				}
				if (gameState == GAME_STATE.WAITING_FOR_USER_INPUT) {
					// either this is a start of a mouse click or a drag
					if (cardPlayField.isMovableCardUnderPoint(mePoint)) {
						cardList = cardPlayField.pickUpCards(mePoint, true);
						if (cardList != null && cardList.size() > 0) {
							sourceMovingCard = new MovingCard(cardList, cardList.get(0).getScreenPosition(),
									mePoint);
							gameState = GAME_STATE.CARD_BEING_GRABBED;
						}
					}
				}
				break;
			case MouseEvent.MOUSE_RELEASED:	// User may have dropped a card
				me = (MouseEvent) event;
				mePoint = me.getPoint();
				if (me.getButton() == MouseEvent.BUTTON1) {
					if (gameState == GAME_STATE.CARD_BEING_GRABBED) {
						// Determine if we can drop a card where the mouse is
						if (!cardPlayField.dropCard(sourceMovingCard.getCardList(), mePoint)) {
							// this was an illegal spot to drop the card
							sourceMovingCard.markAnimationSpot();
							gameState = GAME_STATE.ANIMATE_CARD_RETURN;
						} else {
							canGameAutoMove = true;
							gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
						}
					}
				} 
				
				if (me.getButton() == MouseEvent.BUTTON3 && gameState == GAME_STATE.REVEAL_CARD) {
					// check if user under card, if so remember card to display front and center
					cheatCard = null;
					gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
				}
				
				break;
			case MouseEvent.MOUSE_CLICKED:	// Double click mean user wants to auto-move a card
				me = (MouseEvent) event;
				mePoint = me.getPoint();
				if (me.getClickCount() == 1 && me.getButton() == MouseEvent.BUTTON1) {
					// if user clicks on valid card ...move to sourceMovingCard
					if (gameState == GAME_STATE.WAITING_FOR_USER_INPUT) {
						if (cardPlayField.isSingleClickInValidArea(mePoint) && cardPlayField.isCardUnderPoint(mePoint)) {
							// get the card(s) the user highlighted BUT do not remove from source list
							cardList = cardPlayField.pickUpCards(mePoint, false); 
							if (cardList != null && cardList.size() > 0) {
								Point oldCardPoint = cardList.get(0).getScreenPosition();
								sourceMovingCard = new MovingCard(cardList, oldCardPoint, mePoint);
								// remember this mouse position 
								sourceMovingCard.setOriginalMousePoint(mePoint);
								gameState = GAME_STATE.SOURCE_CARD_SELECTED;
							}
						}
					} else if (gameState == GAME_STATE.SOURCE_CARD_SELECTED) {
						// user selected target
						if (cardPlayField.dropCard(sourceMovingCard.getCardList(), mePoint)) {
							cardList = cardPlayField.pickUpCards(sourceMovingCard.getOriginalMousePoint(), true); 
							if (cardList == null || cardList.size() == 0) {
								System.out.println("can't find this card: " + sourceMovingCard.getCardList().toString());
							}
							// animate the card to the final target location
							canGameAutoMove = true;
							gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
						} else {
							// this was an illegal spot to drop the card
							//sourceMovingCard.markAnimationSpot();
							gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
						}
					}
				} else if (me.getClickCount() == 2 && me.getButton() == MouseEvent.BUTTON1) {
					mePoint = me.getPoint();
					if (gameState == GAME_STATE.WAITING_FOR_USER_INPUT || 
							gameState == GAME_STATE.SOURCE_CARD_SELECTED) {
						if (cardPlayField.isDoubleClickInValidArea(mePoint) && cardPlayField.isCardUnderPoint(mePoint)) {
							cardList = cardPlayField.pickUpCards(mePoint, true); 
							if (cardList != null && cardList.size() == 1) {
								Point theTargetPoint = cardPlayField.determineCardMoveArea(cardList);
								if (theTargetPoint != null) {
									Point oldCardPoint = cardList.get(0).getScreenPosition();
									cardList.get(0).setScreenPosition(theTargetPoint);
									sourceMovingCard = new MovingCard(cardList, oldCardPoint, oldCardPoint);
									sourceMovingCard.markAnimationSpot();
									canGameAutoMove = true;
									gameState = GAME_STATE.ANIMATE_CARD_AUTO_MOVE;
								} else {
									cardPlayField.returnCards(cardList, mePoint);
									gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
								}
							} else {
								gameState = GAME_STATE.WAITING_FOR_USER_INPUT;
							}
						}
						
					}
				}
				break;
			default:
				//System.out.println("Event occurred that we are not handling. " + event.toString());
		}
		
	}



}
